
const B3MeasureExt = () => {

'use strict';

var avem = AutodeskNamespace('Autodesk.Viewing.Extensions.B3LMeasure'),
    av = Autodesk.Viewing,
    avu = av.UI;

var NONE = 0;
var MEASURE_TOOL = 1;

var DEFAULT_MEASUREMENT_TYPE = avem.MEASUREMENT_POINT;

/**
 * Extension used to support distance and angle measure for 2d and 3d models.
 * @constructor
 * @tutorial feature_measure
 * @param {Autodesk.Viewing.Viewer3D} viewer - Viewer to be extended.
 * @param {object} options - An optional dictionary of options for this extension.
 * @memberof Autodesk.Viewing.Extensions.B3LMeasure
 * @alias Autodesk.Viewing.Extensions.B3LMeasure.MeasureExtension
 * @category Extensions
*/
var MeasureExtension = function(viewer, options) {
    Autodesk.Viewing.Extension.call(this, viewer, options);
};

MeasureExtension.prototype = Object.create(Autodesk.Viewing.Extension.prototype);
MeasureExtension.prototype.constructor = MeasureExtension;


// MeasureExtension.prototype.onToolbarCreated = function() {
//     this.viewer.removeEventListener(av.TOOLBAR_CREATED_EVENT, this.bindedOnToolbarCreated);
//     this.bindedOnToolbarCreated = null;
//     this.createUI();
// };

/**
 * Load the measure extension.
 * @returns {boolean} True if measure extension is loaded successfully.
*/
MeasureExtension.prototype.load = function() {
    console.log("MeasureExtension Load called");

    var self   = this;
    var viewer = this.viewer;
    this.hasUI = Autodesk.Viewing.Private.GuiViewer3D && viewer instanceof Autodesk.Viewing.Private.GuiViewer3D;

    this.escapeHotkeyId = 'Autodesk.B3LMeasure.Hotkeys.Escape';

    // Register the Measure tool
    if (!viewer.toolController){
        return false;
    }

    this.options = this.options || {};
    var measureToolOptions = {};

    measureToolOptions.onCloseCallback = function() {
        self.enableMeasureTool(false);
    };


    // Shared State with measureTool & calibrationTool.
    this.sharedMeasureConfig = {
        units: viewer.model.getDisplayUnit(),
        precision: viewer.model && viewer.model.is2d() ? 3 : 1,
        calibrationFactor: this.options.calibrationFactor
    };

	measureToolOptions.snapperOptions = this.options.snapperOptions;

    this.forceCalibrate = this.options.forceCalibrate;

    this.isCalibrated = (this.options.calibrationFactor != null);

    this.snapper = new Autodesk.Viewing.Extensions.B3LMeasure.Snapper(viewer);
    viewer.toolController.registerTool(this.snapper);

    this.measureTool = new avem.MeasureTool(viewer, measureToolOptions, this.sharedMeasureConfig, this.snapper);
    viewer.toolController.registerTool(this.measureTool);

    this.magnifyingGlass = new Autodesk.Viewing.Extensions.B3LMeasure.MagnifyingGlass(viewer);
    viewer.toolController.registerTool(this.magnifyingGlass);

    this.onMeasurementChanged = function(event) {
        var type = event.data;
        self.changeMeasurementType(type);
    };

    viewer.addEventListener(avem.MEASUREMENT_CHANGED_EVENT, this.onMeasurementChanged);

//    if (this.viewer.toolbar) {
//        this.createUI();
//    } else {
//        this.bindedOnToolbarCreated = this.onToolbarCreated.bind(this);
//        this.viewer.addEventListener(av.TOOLBAR_CREATED_EVENT, this.bindedOnToolbarCreated);
//    }

    return true;
};

/**
 * Unload the measure extension.
 * @returns {boolean} True if measure extension is unloaded successfully.
*/
MeasureExtension.prototype.unload = function () {
    var viewer = this.viewer;

    // Remove the ui from the viewer.
    this.destroyUI();
    if (this.bindedOnToolbarCreated) {
        this.viewer.removeEventListener(av.TOOLBAR_CREATED_EVENT, this.bindedOnToolbarCreated);
        this.bindedOnToolbarCreated = null;
    }

    viewer.removeEventListener(avem.MEASUREMENT_CHANGED_EVENT, this.onMeasurementChanged);

    viewer.toolController.deregisterTool(this.snapper);
    this.snapper = null;

    viewer.toolController.deregisterTool(this.measureTool);
    this.measureTool = null;

    viewer.toolController.deregisterTool(this.magnifyingGlass);
    this.magnifyingGlass = null;

    return true;
};

/**
 * Whether the measure tool is currently active.
 * @returns {boolean}
 */
MeasureExtension.prototype.isActive = function() {
    return this.measureTool.isActive();
};

/**
 * Enable/disable the measure tool.
 * @param {boolean} active - True to activate, false to deactivate.
 * @returns {boolean} True if a change in activeness occurred.
 */
MeasureExtension.prototype.setActive = function(active) {
    return this.enableMeasureTool(active);
};

/**
 * Toggles activeness of the measure tool.
 * @return {boolean} Whether the tool is active.
 */
MeasureExtension.prototype.toggle = function() {
    if (this.isActive()) {
        this.enableMeasureTool(false);
    } else {
        this.enableMeasureTool(true);
    }
    return this.isActive();
};

/**
 * Get the current measurement in the measure tool.
 * @param {string} unitType - Optional measure unit: "decimal-ft", "ft", "ft-and-decimal-in", "decimal-in", "fractional-in", "m", "cm", "mm" or "m-and-cm".
 * @param {number} precision - Optional measure precision index (0: 0, 1: 0.1, 2: 0.01, 3: 0.001, 4: 0.0001, 5: 0.00001).
 * When units type is "ft", "in" or "fractional-in", then the precisions are 0: 1, 1: 1/2, 2: 1/4, 3: 1/8, 4: 1/16, 5: 1/32, 6: 1/64.
 * @returns {object|null} Object with properties of the current measurement, or null.
 */
MeasureExtension.prototype.getMeasurement = function(unitType, precision) {
    var measurement = null;
    if (this.measureTool.isActive()) {
        measurement = this.measureTool.getMeasurement(unitType, precision);
    }
    return measurement;
};

/**
 * Get all available units in measure tool.
 * @returns {object[]} Array of all available units.
*/
MeasureExtension.prototype.getUnitOptions = function() {
    var units = [
        { name: 'Unknown', type: '' },
        { name: 'Decimal feet', type: 'decimal-ft' },
        { name: 'Feet and fractional inches', type: 'ft' },
        { name: 'Feet and decimal inches', type: 'ft-and-decimal-in' },
        { name: 'Decimal inches', type: 'decimal-in' },
        { name: 'Fractional inches', type: 'fractional-in' },
        { name: 'Meters', type: 'm' },
        { name: 'Centimeters', type: 'cm' },
        { name: 'Millimeters', type: 'mm' },
        { name: 'Meters and centimeters', type: 'm-and-cm' }
    ];

    return units;
};

/**
 * Get all available precisions in measure tool.
 * @param {boolean} isFractional - Set true to get fractional precisions.
 * @return {string[]} List of all available precisions.
*/
MeasureExtension.prototype.getPrecisionOptions = function(isFractional) {

    var precisions;

    if (isFractional)
        precisions = ['1', '1/2', '1/4', '1/8', '1/16', '1/32', '1/64'];
    else
        precisions = ['0', '0.1', '0.01', '0.001', '0.0001', '0.00001'];

    return precisions;
};

/**
 * Get the default measure unit in measure tool.
 * @returns {string} The default measure unit.
*/
MeasureExtension.prototype.getDefaultUnit = function() {
    var unit = this.viewer.model.getDisplayUnit();

    return unit;
};

/**
 * Enable/disable the measure tool.
 * @param {boolean} enable - True to enable, false to disable.
 * @returns {boolean} True if the tool state was changed.
 * @private
 */
MeasureExtension.prototype.enableMeasureTool = function(enable, measurementType) {
    if (measurementType === avem.MEASUREMENT_AREA && this.viewer.model && !this.viewer.model.is2d()) {
        return false;
    }

    var toolController = this.viewer.toolController,
        isActive = (this.selectedTool === MEASURE_TOOL);

    this.forceCalibrate |= this.viewer.model.getData().isLeaflet;

    if (!measurementType) {
        measurementType = DEFAULT_MEASUREMENT_TYPE;
    }

    if (enable && !isActive) {

        // Fetch topology when opening Measure tool for the first time.
        this.checkAndFetchTopology(toolController.getTool('B3Lmeasure'));
        toolController.activateTool("B3Lmeasure");
        this.selectedTool = MEASURE_TOOL;
        this.changeMeasurementType(measurementType);
        return true;
    } else if (!enable && isActive) {
        if (this.measureTool.isActive()) {
            toolController.deactivateTool("B3Lmeasure");

        }
        this.selectedTool = NONE;
        return true;
    } else if (enable && isActive) {
        this.changeMeasurementType(measurementType);
        return true;
    }
    return false;
};

MeasureExtension.prototype.changeMeasurementType = function(measurementType) {
    this.measureTool.changeMeasurementType(measurementType);
};

/**
 * @private
 */
MeasureExtension.prototype.enterMeasurementMode = function() {

    if (this._measurementMode) return;
    this._measurementMode = true;

    var toolbar = this.viewer.getToolbar(false);    
    var viewerToolbarContainer = toolbar.container;
    var viewerContainerChildrenCount = viewerToolbarContainer.children.length;
    for(var i = 0; i < viewerContainerChildrenCount; ++i) {
        viewerToolbarContainer.children[i].style.display = "none";
    }

    this.navigationControls = toolbar.getControl(Autodesk.Viewing.TOOLBAR.NAVTOOLSID);
    this.navigationControls.setVisible(true);
    this.navigationControls.container.style.display = '';

    this.measureControls = toolbar.getControl(Autodesk.Viewing.TOOLBAR.MEASURETOOLSID);
    this.measureControls.setVisible(true);
    this.measureControls.container.style.display = '';
    
    var modelTools = toolbar.getControl(av.TOOLBAR.MODELTOOLSID);
    var measureButtonId = this.measurementToolbarButton.getId();
    this.measurementToolbarButton.index = modelTools.indexOf(measureButtonId);
    modelTools.removeControl(measureButtonId);

    if (this.viewer.centerToolBar) {
        this.viewer.centerToolBar();
    }

    this.enableMeasureTool(true, DEFAULT_MEASUREMENT_TYPE);
};

/**
 * @private
 */
MeasureExtension.prototype.exitMeasurementMode = function() {
    if (!this._measurementMode) return;
    this._measurementMode = false;
};

/**
 * Create measure button in toolbar.
 * @private
*/
MeasureExtension.prototype.createUI = function()
{
    var self   = this;
    var viewer = this.viewer;

    // // Escape hotkey to exit tool.
    // //
    // var hotkeys = [{
    //     keycodes: [
    //         av.theHotkeyManager.KEYCODES.ESCAPE
    //     ],
    //     onRelease: function () {
    //         if (self._measurementMode) {
    //             self.exitMeasurementMode();
    //             return true;
    //         }
    //     }
    // }];
    //av.theHotkeyManager.pushHotkeys(this.escapeHotkeyId, hotkeys);

    // Finally
    this.uiCreated = true;
};


/**
 * @private
 */
MeasureExtension.prototype.checkAndFetchTopology = function(tool) {

    if (!this.uiCreated || !this.viewer.model.is3d()) {
        tool && tool.setNoTopology();
        return;
    }

    if (this.viewer.modelHasTopology()){
        tool && tool.setTopologyAvailable();
        return;
    }

    // Fetch topology from backend.
    tool && tool.setFetchingTopology();
    this.viewer.model.fetchTopology()
    .then(function(topoData){
        tool && tool.setTopologyAvailable();
    })
    .catch(function(err){
        avp.logger.log(err); // No topology
        tool && tool.setNoTopology();
    });
};

/**
 * Destroy measure button in toolbar.
 * @private
*/
MeasureExtension.prototype.destroyUI = function()
{
    var viewer = this.viewer;    

    var toolbar = viewer && viewer.getToolbar && viewer.getToolbar(false);
    if (toolbar) {
        var modelTools = toolbar.getControl(av.TOOLBAR.MODELTOOLSID);
        if (modelTools) {
            var submenu = null;
            
            if (this.measurementToolbarButton) {
                submenu = modelTools.getControl("toolbar-B3LinspectSubMenu");
                if (submenu) {
                    submenu.removeControl(this.measurementToolbarButton.getId());
                } else {
                    modelTools.removeControl(this.measurementToolbarButton.getId());
                }
            }

            this.measurementToolbarButton = null;
        }
    }

    av.theHotkeyManager.popHotkeys(this.escapeHotkeyId);
};

MeasureExtension.prototype.setUnits = function(units) {
    this.measureTool.setUnits(units);
};

MeasureExtension.prototype.getUnits = function() {
    return this.measureTool.getUnits();
};

MeasureExtension.prototype.setPrecision = function(precision) {
    this.measureTool.setPrecision(precision);
};

MeasureExtension.prototype.getPrecision = function() {
    return this.measureTool.getPrecision();
};

MeasureExtension.prototype.getCalibrationFactor = function() {
    return  this.sharedMeasureConfig.calibrationFactor;
};

MeasureExtension.prototype.deleteCurrentMeasurement = function() {
    this.measureTool.deleteCurrentMeasurement();
};

avem.MeasureExtension = MeasureExtension;
av.theExtensionManager.registerExtension('Autodesk.B3LMeasure', MeasureExtension);


}
