(function(){ 'use strict';

    // Measurement helper functions //
    var MeasureCommon = AutodeskNamespace("Autodesk.Viewing.Extensions.B3LMeasure.Functions");
    var avem = Autodesk.Viewing.Extensions.B3LMeasure;
    
    avem.SNAP_VERTEX = 0;
    avem.SNAP_MIDPOINT = 1;
    avem.SNAP_CIRCLE_CENTER = 2;
    avem.SNAP_EDGE = 3;
    avem.SNAP_FACE = 4;
    avem.SNAP_CIRCULARARC = 5;
    avem.SNAP_CURVEDEDGE = 6;
    avem.SNAP_CURVEDFACE = 7;

    avem.MEASUREMENT_POINT = 0;
    avem.MEASUREMENT_DISTANCE = 1; // Measurement from point to point, not matter what geometry it is.
    avem.MEASUREMENT_ANGLE = 2;
    avem.MEASUREMENT_AREA = 3;

    avem.MEASUREMENT_CHANGED_EVENT = 'measurement-changed';
    
    avem.EPSILON = 0.0001;

    MeasureCommon.getSnapResultPosition = function(pick, viewer) {
        if (!pick) {
            return null;
        }

        if (pick.isPerpendicular) {
            return pick.intersectPoint;
        }

        switch (pick.geomType) {
            case avem.SNAP_VERTEX:
            case avem.SNAP_MIDPOINT:
            case avem.SNAP_CIRCLE_CENTER:
                return pick.getGeometry();

            case avem.SNAP_EDGE:
                var eps = MeasureCommon.getEndPointsInEdge(pick.getGeometry());
                var p1 = eps[0].clone();
                var p2 = eps[1].clone();
                return MeasureCommon.nearestPointInPointToLine(pick.intersectPoint,p1, p2);

            case avem.SNAP_FACE:
                return MeasureCommon.nearestPointInPointToPlane(pick.intersectPoint, pick.getGeometry().vertices[0], pick.faceNormal);

            case avem.SNAP_CIRCULARARC:
                if (viewer && viewer.model && viewer.model.is2d()) {
                    var point = MeasureCommon.nearestVertexInVertexToEdge(pick.intersectPoint, pick.getGeometry());
                    pick.geomType = avem.SNAP_VERTEX;
                    pick.geomVertex = point;
                    return point;

                } else {
                    // For 3D models, currently we don't have the center geometry of the circle. 
                    // So the only way to select the center is by selecting the perimeter.
                    return pick.circularArcCenter;    
                }

            case avem.SNAP_CURVEDEDGE:
                return MeasureCommon.nearestVertexInVertexToEdge(pick.intersectPoint, pick.getGeometry());

            case avem.SNAP_CURVEDFACE:
                return pick.intersectPoint;

            default:
                return null;
        }
    };

    MeasureCommon.correctPerpendicularPicks = function(firstPick, secondPick, viewer, snapper) {

        if (!firstPick || !secondPick || !firstPick.getGeometry() || !secondPick.getGeometry()) {
            return false;
        }

        var start = MeasureCommon.getSnapResultPosition(firstPick, viewer);

        if (snapper && viewer) {

            // Simple _ to Edge - Snap the second pick when it's 90 degrees.
            if (secondPick.geomType === avem.SNAP_EDGE) {
                var v2 = new THREE.Vector3();
                var v3 = new THREE.Vector3();
                
                var secondEdge = secondPick.getGeometry();

                v2.subVectors(secondPick.intersectPoint, start).normalize(); // rubberband vector
                v3.subVectors(secondEdge.vertices[0], secondEdge.vertices[1]).normalize();

                if(MeasureCommon.isPerpendicular(v2, v3, 0.05)) {
                    
                    var newPoint = MeasureCommon.nearestPointInPointToSegment(start, secondEdge.vertices[0], secondEdge.vertices[1], true);
                    
                    if (newPoint) {
                        if (snapper.onMouseMove(MeasureCommon.project(newPoint, viewer))) {
                            snapper.setPerpendicular(true);
                        }
                        
                        secondPick.geomVertex = newPoint;
                        secondPick.intersectPoint = newPoint;
                        return true;
                    }
                }
            }

            // Simple _ to Face - Snap the second pick when it's 90 degrees.
            else if (secondPick.geomType === avem.SNAP_FACE) {

                var v = new THREE.Vector3();
                
                var secondFace = secondPick.getGeometry();

                v.subVectors(secondPick.intersectPoint, start).normalize(); // rubberband vector

                if(MeasureCommon.isParallel(secondPick.faceNormal, v, 0.05)) {
                    
                    var newPoint = MeasureCommon.nearestPointInPointToPlane(start, secondFace.vertices[0], secondPick.faceNormal);
                    if (snapper.onMouseMove(MeasureCommon.project(newPoint, viewer))) {
                        snapper.setPerpendicular(true);
                    }

                    secondPick.geomVertex = newPoint;
                    secondPick.intersectPoint = newPoint;
                    return true;
                }
            }
        }
    };

    MeasureCommon.calculatePoint = function(firstPick, viewer) {

        if (!firstPick || !firstPick.getGeometry()) {
            return null;
        }

        var ep1 = MeasureCommon.getSnapResultPosition(firstPick, viewer);

        if (!ep1) {
            return null;
        }

        var XYZ, X, Y, Z;

        // Convert coords when in 2D
        if (viewer.model.is2d()) {
            ep1 = ep1.clone();
            viewer.model.pageToModel(ep1,firstPick.viewportIndex2d);
        }

        // Calculation for 3D models and 2D models without resolution limits
        return {
            X: ep1.x,
            Y: ep1.y,
            Z: ep1.z,
            type: avem.MEASUREMENT_POINT
        };
    };
    
    MeasureCommon.calculateDistance = function(firstPick, secondPick, dPrecision, viewer) {

        if (!firstPick || !secondPick || !firstPick.getGeometry() || !secondPick.getGeometry()) {
            return null;
        }

        var ep1 = MeasureCommon.getSnapResultPosition(firstPick, viewer);
        var ep2 = MeasureCommon.getSnapResultPosition(secondPick, viewer);

        if (!ep1 || !ep2) {
            return null;
        }

        if (MeasureCommon.isEqualVectors(ep1, ep2, avem.EPSILON)) {
            return null;
        }

        var distanceXYZ, distanceX, distanceY, distanceZ;

        // Convert coords when in 2D
        if (viewer.model.is2d()) {
            ep1 = ep1.clone();
            ep2 = ep2.clone();
            viewer.model.pageToModel(ep1, ep2, firstPick.viewportIndex2d);
        }

        // Include resolution limits for high precision 2D-measurements, where available 
        if (dPrecision) {

            var n = Math.log(2.0 * dPrecision) / Math.log(10.0);
            var nd = Math.floor(n);

            // Increase one decimal to ensure being covered
            if (1. < (2.0 * dPrecision / Math.pow(10., nd)))
                nd++;
                
            dPrecision = Math.pow(10., nd);

            // Adjust the distances aligned by the precision
            var measurementDistance = ep1.distanceTo(ep2);
            var m = measurementDistance / dPrecision;
            distanceXYZ = Math.floor(m + 0.5) * dPrecision;
            
            measurementDistance = Math.abs(ep1.x - ep2.x);
            m = measurementDistance / dPrecision;
            distanceX = Math.floor(m + 0.5) * dPrecision;

            measurementDistance = Math.abs(ep1.y - ep2.y);
            m = measurementDistance / dPrecision;
            distanceY = Math.floor(m + 0.5) * dPrecision;

            return {
                distanceXYZ: distanceXYZ,
                distanceX: distanceX,
                distanceY: distanceY,
                distanceZ: 0,
                type: avem.MEASUREMENT_DISTANCE
            };
        }

        // Calculation for 3D models and 2D models without resolution limits
        distanceXYZ = ep1.distanceTo(ep2);
        distanceX = Math.abs(ep1.x - ep2.x);
        distanceY = Math.abs(ep1.y - ep2.y);
        distanceZ = Math.abs(ep1.z - ep2.z);
        return {
            distanceXYZ: distanceXYZ,
            distanceX: distanceX,
            distanceY: distanceY,
            distanceZ: distanceZ,
            type: avem.MEASUREMENT_DISTANCE
        };
    };

    MeasureCommon.getDpiPrecision = function(model, viewportIndex) {

        // Only do this for 2D models.
        if (model.is3d()) {
            return 0;
        }

        // Include resolution limits for high precision 2D-measurements, where available (TREX-575)
        var page_width = model.getMetadata('page_dimensions', 'page_width', null);
        var logical_width = model.getMetadata('page_dimensions', 'logical_width', null);
        var page_height = model.getMetadata('page_dimensions', 'page_height', null);
        var logical_height = model.getMetadata('page_dimensions', 'logical_height', null);
        if (!page_width || !logical_width || !page_height || !logical_height) {
            return 0;
        }

        // Retrieve the inverse DPI
        var invdpix = page_width / logical_width;
        var invdpiy = page_height / logical_height;
        
        // Calculate the graininess in model units
        var p1 = new THREE.Vector3(0.0, 0.0, 0.0);
        var p2 = new THREE.Vector3(invdpix, invdpiy, 0.0);
        model.pageToModel(p1, p2, viewportIndex);
        var dPrecision = p1.distanceTo(p2);

        return dPrecision;
    };

    MeasureCommon.isContainsEqualVectors = function(points) {
        for (var i = 0; i < points.length; i++) {
            for (var j = 0; j < points.length; j++) {
                if (i !== j && MeasureCommon.isEqualVectors(points[i], points[j], avem.EPSILON)) {
                    return true;
                }
            }
        }

        return false;
    };

    MeasureCommon.calculateAngle = function(picks, viewer) {
        var points = [];

        for (var key in picks) {
            if (picks.hasOwnProperty(key)) {
                var point = MeasureCommon.getSnapResultPosition(picks[key], viewer);
                if (point) {
                    points.push(point);    
                }
            }            
        }

        if (points.length !== 3 || MeasureCommon.isContainsEqualVectors(points)) {
            return null;
        }


        var v1 = new THREE.Vector3();
        var v2 = new THREE.Vector3();

        v1.subVectors(points[0], points[1]);
        v2.subVectors(points[2], points[1]);

        return MeasureCommon.angleVectorToVector(v1, v2);
    };
    
    MeasureCommon.calculateArea = function(picks, viewer) {
        var points = [];

        for (var key in picks) {
            if (picks.hasOwnProperty(key)) {
                var point = MeasureCommon.getSnapResultPosition(picks[key], viewer);
                if (point) {
                    points.push(point.clone());    
                }
            }            
        }

        var firstPoint = MeasureCommon.getSnapResultPosition(picks[1], viewer);
        if (firstPoint) {
            points.push(firstPoint.clone());
        }

        for (var i = 0; i < points.length; i+=2) {
            viewer.model.pageToModel(points[i], points[i + 1], picks[1].viewportIndex2d);
        }

        var sum1 = 0;
        var sum2 = 0;

        for (var i = 0; i < points.length - 1; i++) {
            sum1 += points[i].x * points[i+1].y;
            sum2 += points[i].y * points[i+1].x;
        }

        var area = Math.abs((sum1 - sum2) / 2);
        return area;
        
    };

    /**
     * The main function for this file, which calculates a measurement result (either distance
     * or angle) from a given measurement.
     */
    MeasureCommon.computeResult = function(picks, measurementType, viewer) {

        switch (measurementType) {
            case avem.MEASUREMENT_POINT:
                return MeasureCommon.calculatePoint(picks[1], viewer);

            case avem.MEASUREMENT_DISTANCE:
                var firstPick = picks[1];
                var secondPick = picks[2];
                var dPrecision = MeasureCommon.getDpiPrecision(viewer.model, firstPick.viewportIndex2d);

                return MeasureCommon.calculateDistance(firstPick, secondPick, dPrecision, viewer);
            
            case avem.MEASUREMENT_ANGLE:
                var angle = MeasureCommon.calculateAngle(picks, viewer);
                return angle ? {
                        angle: angle,
                        type: avem.MEASUREMENT_ANGLE
                    } : null;

            case avem.MEASUREMENT_AREA:
                return {
                    area: MeasureCommon.calculateArea(picks, viewer),
                    type: avem.MEASUREMENT_AREA
                };

            default:
                return null;

        }
    };

    MeasureCommon.getFaceArea = function(viewer, measurement, face, units, precision, calibrationFactor) {

        var area = 0;
        var vertices = face.vertices;
        var V1 = new THREE.Vector3();
        var V2 = new THREE.Vector3();

        for (var i = 0; i < vertices.length; i += 3) {

            V1.subVectors(vertices[i + 1], vertices[i]);
            V2.subVectors(vertices[i + 2], vertices[i]);

            area += V1.length() * V2.length() * Math.sin(V1.angleTo(V2)) / 2;
        }

        area = Autodesk.Viewing.Private.convertUnits(viewer.model.getUnitString(), units, calibrationFactor, area, 'square');

        if (units) {

            return Autodesk.Viewing.Private.formatValueWithUnits(area, units+'^2', 3, precision);
        }
        else {

            return Autodesk.Viewing.Private.formatValueWithUnits(area, null, 3, precision);
        }

    };

    MeasureCommon.getCircularArcRadius = function(viewer, measurement, edge, units, precision, calibrationFactor) {

        var radius = edge.radius;

        if (radius) {
            if (viewer.model.is2d()) {
                var pt1 = edge.center.clone();
                var pt2 = edge.vertices[0].clone();
                viewer.model.pageToModel(pt1, pt2, measurement.getPick(1).viewportIndex2d);
                radius = pt1.distanceTo(pt2);
            }

            radius = Autodesk.Viewing.Private.convertUnits(viewer.model.getUnitString(), units, calibrationFactor, radius);
            return Autodesk.Viewing.Private.formatValueWithUnits(radius, units, 3, precision);
        }
    };

    MeasureCommon.project = function(point, viewer, offset) {
        var camera = viewer.navigation.getCamera(),
            containerBounds = viewer.navigation.getScreenViewport(),
            p = new THREE.Vector3(point.x, point.y, point.z);

        p = p.project(camera);

        offset = offset || 0;

        return new THREE.Vector2(Math.round(( p.x + 1) / 2 * containerBounds.width) + offset,
                Math.round((-p.y + 1) / 2 * containerBounds.height) + offset);
    };

    MeasureCommon.inverseProject = function (point, viewer) {

        var camera = viewer.navigation.getCamera(),
            containerBounds = viewer.navigation.getScreenViewport(),
            p = new THREE.Vector3();

        p.x = point.x / containerBounds.width * 2 - 1;
        p.y = -(point.y / containerBounds.height * 2 - 1);
        p.z = 0;

        p = p.unproject(camera);

        return p;
    };

    //***** Helper functions for calculations without state: ***** //
    

    // Get the nearest point on the line from point to line
    // inLine - whether to force the result to be inside the given line or not.
    MeasureCommon.nearestPointInPointToLine = function(point, lineStart, lineEnd) {

        var X0 = new THREE.Vector3();
        var X1 = new THREE.Vector3();
        var nearestPoint;
        var param;

        X0.subVectors(lineStart, point);
        X1.subVectors(lineEnd, lineStart);
        param = X0.dot(X1);
        X0.subVectors(lineEnd, lineStart);
        param = -param / X0.dot(X0);

        X0.subVectors(lineEnd, lineStart);
        X0.multiplyScalar(param);
        nearestPoint = X0.add(lineStart);    

        return nearestPoint;
    };

    // Get the nearest point on the line segment from point to line segment
    MeasureCommon.nearestPointInPointToSegment = function (point, lineStart, lineEnd, forcePerpendicular) {

        var X0 = new THREE.Vector3();
        var X1 = new THREE.Vector3();
        var nearestPoint;
        var param;

        X0.subVectors(lineStart, point);
        X1.subVectors(lineEnd, lineStart);
        param = X0.dot(X1);
        X0.subVectors(lineEnd, lineStart);
        param = -param / X0.dot(X0);

        if (param < 0) {
            if (forcePerpendicular) {
                nearestPoint = null;
            } else {
                nearestPoint = lineStart;    
            }
        }
        else if (param > 1) {
            if (forcePerpendicular) {
                nearestPoint = null;
            } else {
                nearestPoint = lineEnd;    
            }
        }
        else {
            X0.subVectors(lineEnd, lineStart);
            X0.multiplyScalar(param);
            nearestPoint = X0.add(lineStart);
        }

        return nearestPoint;
    };
        
        

    MeasureCommon.isEqualVectors = function (v1, v2, precision) {
        if (!v1 || !v2) {
            return false;
        }

        if (Math.abs(v1.x - v2.x) <= precision && Math.abs(v1.y - v2.y) <= precision && Math.abs(v1.z - v2.z) <= precision) {
            return true;
        }
        return false;
    };

    MeasureCommon.isInverseVectors = function (v1, v2, precision) {
        if (Math.abs(v1.x + v2.x) <= precision && Math.abs(v1.y + v2.y) <= precision && Math.abs(v1.z + v2.z) <= precision) {
            return true;
        }
        return false;
    };

    MeasureCommon.getEndPointsInEdge = function(edge) {

        var vertices = edge.vertices;
        var endPoints = [];

        for (var i = 0; i < vertices.length; ++i) {

            var duplicate = false;

            for (var j = 0; j < vertices.length; ++j) {

                if (j !== i && vertices[j].equals(vertices[i])) {

                    duplicate = true;
                    break;
                }
            }

            if (!duplicate) {

                endPoints.push(vertices[i]);

            }
        }

        return endPoints;
    };

    MeasureCommon.angleVectorToVector = function(v1, v2) {
        return v1.angleTo(v2) * 180 / Math.PI;
    };

    // Get the two nearest endpoints between two line segments
    MeasureCommon.nearestPointsInSegmentToSegment = function (p1, p2, p3, p4) {

        
        var u = new THREE.Vector3();
        var v = new THREE.Vector3();
        var w = new THREE.Vector3();

        u.subVectors(p2, p1);
        v.subVectors(p4, p3);
        w.subVectors(p1, p3);

        var a = u.dot(u);
        var b = u.dot(v);
        var c = v.dot(v);
        var d = u.dot(w);
        var e = v.dot(w);
        var D = a * c - b * b;
        var sc, sN, sD = D;
        var tc, tN, tD = D;

        // Compute the line parameters of the two closest points
        if (D < avem.EPSILON) { // the lines are almost parallel
            sN = 0.0;  // for using point p1 on segment p1p2
            sD = 1.0;  // to prevent possible division by 0.0 later
            tN = e;
            tD = c;
        }
        else {  // get the closest points on the infinite lines
            sN = b * e - c * d;
            tN = a * e - b * d;
            if (sN < 0.0) {  // sc < 0 => the s = 0 is visible
                sN = 0.0;
                tN = e;
                tD = c;
            }
            else if (sN > sD) {  // sc > 1 => the s = 1 edge is visible
                sN = sD;
                tN = e + b;
                tD = c;
            }
        }

        if (tN < 0.0) {  // tc < 0 => the t = 0 edge is visible
            tN = 0.0;
            // recompute sc for this edge
            if (-d < 0.0)
                sN = 0.0;
            else if (-d > a)
                sN = sD;
            else {
                sN = -d;
                sD = a;
            }
        }
        else if (tN > tD) {  // tc > 1 => the t = 1 edge is visible
            tN = tD;
            // recompute sc for this edge
            if ((-d + b) < 0.0)
                sN = 0;
            else if ((-d + b) > a)
                sN = sD;
            else {
                sN = -d + b;
                sD = a;
            }
        }

        // finally do the division to get sc and tc
        sc = Math.abs(sN) < avem.EPSILON ? 0.0 : sN / sD;
        tc = Math.abs(tN) < avem.EPSILON ? 0.0 : tN / tD;

        // get the difference of the two closest points
        u.multiplyScalar(sc);
        v.multiplyScalar(tc);
        w.add(u);
        w.sub(v);

        //return w.length();

        u.add(p1);
        v.add(p3);
        return [u, v];
    };


    // Get the nearest point on edge from point to edge
    MeasureCommon.nearestPointInPointToEdge = function(edge, point) {

        var vertices = edge.vertices;
        var minDist = Number.MAX_VALUE;
        var nearestPoint = null;

        for (var i = 0; i < vertices.length; i += 2) {

            var nP = MeasureCommon.nearestPointInPointToSegment(point, vertices[i], vertices[i + 1]);
            var dist = point.distanceTo(nP);
            if (dist < minDist) {
                minDist = dist;
                nearestPoint = nP;
            }
        }

        return nearestPoint;
    };

    // Find the nearest point from point to plane
    MeasureCommon.nearestPointInPointToPlane = function(p1, p2, n) {

        var nearestPoint = new THREE.Vector3();
        var norm = n.clone();
        var X0 = new THREE.Vector3();
        X0.subVectors(p1, p2);

        var sn = -norm.dot(X0);
        var sd = norm.dot(norm);
        var sb = sn / sd;

        nearestPoint.addVectors(p1, norm.multiplyScalar(sb));
        return nearestPoint;
    };

    // Find the intersection point of two nonparallel lines
    MeasureCommon.intersectLineToLine = function(p1, v1, p2, v2) {

        var X0 = new THREE.Vector3();
        var X1 = new THREE.Vector3();

        X0.subVectors(p2, p1);
        X0.cross(v2);
        X1.crossVectors(v1, v2);

        var scalar = X0.divide(X1);

        X1 = v1.clone();
        X1.multiplyScalar(scalar);
        X0.addVectors(p1, X1);

        return X0;
    };



    // Returns true if v1 an v2 are parallel
    MeasureCommon.isParallel = function(v1, v2, precision) {
        precision = precision ? precision : avem.EPSILON;
        return 1 - Math.abs(v1.dot(v2)) < precision;
    };

    MeasureCommon.isPerpendicular = function(v1, v2, precision) {
        precision = precision ? precision : avem.EPSILON;
        return Math.abs(v1.dot(v2)) < precision;
    };


    // Find the intersection of two nonparallel planes
    MeasureCommon.intersectPlaneToPlane = function(p1, n1, p2, n2) {

        var u = new THREE.Vector3();
        u.crossVectors(n1, n2);
        var ax = (u.x >= 0 ? u.x : -u.x);
        var ay = (u.y >= 0 ? u.y : -u.y);
        var az = (u.z >= 0 ? u.z : -u.z);

        var maxc;  // max coordinate
        if (ax > ay) {
            if (ax > az)
                maxc = 1;
            else
                maxc = 3;
        }
        else {
            if (ay > az)
                maxc = 2;
            else maxc = 3;
        }

        var iP = new THREE.Vector3(); // intersect point
        var d1, d2;
        d1 = -n1.dot(p1);
        d2 = -n2.dot(p2);

        switch (maxc) {

            case 1:  // intersect with x = 0
                iP.x = 0;
                if (u.x !== 0) {
                    iP.y = (d2 * n1.z - d1 * n2.z) / u.x;
                    iP.z = (d1 * n2.y - d2 * n1.y) / u.x;
                }
                else {
                    iP.y = - (d2 * n1.z) / (n1.z * n2.y);
                    iP.z = - (d1 * n2.y) / (n1.z * n2.y);
                }
                break;
            case 2:
                iP.y = 0;
                if (u.y !== 0) {
                    iP.x = (d1 * n2.z - d2 * n1.z) / u.y;
                    iP.z = (d2 * n1.x - d1 * n2.x) / u.y;
                }
                else {
                    iP.x = - (d1 * n2.z) / (n1.x * n2.z);
                    iP.z = - (d2 * n1.x) / (n1.x * n2.z);
                }
                break;
            case 3:
                iP.z = 0;
                if (u.z !== 0) {
                    iP.x = (d2 * n1.y - d1 * n2.y) / u.z;
                    iP.y = (d1 * n2.x - d2 * n1.x) / u.z;
                }
                else {
                    iP.x = - (d2 * n1.y) / (n1.y * n2.x);
                    iP.y = - (d1 * n2.x) / (n1.y * n2.x);
                }
                break;
        }

        var iP2 = new THREE.Vector3();
        iP2.addVectors(iP, u.multiplyScalar(100));

        var vP1 = MeasureCommon.nearestPointInPointToLine(p1, iP, iP2);
        var vP2 = MeasureCommon.nearestPointInPointToLine(p2, iP, iP2);

        return [vP1, vP2];

    };

    // Find the nearest point from point to triangle
    MeasureCommon.nearestPointInPointToTriangle = function(point, a, b, c) {

        var nearestPoint;
        var minDist = Number.MAX_VALUE;

        nearestPoint = MeasureCommon.pointProjectsInTriangle(point, a, b, c);
        if (nearestPoint) {
            return nearestPoint;
        }

        var p = MeasureCommon.nearestPointInPointToSegment(point, a, b);
        if (point.distanceTo(p) < minDist) {
            minDist = point.distanceTo(p);
            nearestPoint = p.clone();
        }

        p = MeasureCommon.nearestPointInPointToSegment(point, a, c);
        if (point.distanceTo(p) < minDist) {
            minDist = point.distanceTo(p);
            nearestPoint = p.clone();
        }

        p = MeasureCommon.nearestPointInPointToSegment(point, b, c);
        if (point.distanceTo(p) < minDist) {

            nearestPoint = p.clone();
        }

        return nearestPoint;
    };

    MeasureCommon.pointProjectsInTriangle = function(point, a, b, c) {

        var u = new THREE.Vector3();
        var v = new THREE.Vector3();
        var w = new THREE.Vector3();
        var n = new THREE.Vector3();

        u.subVectors(b, a);
        v.subVectors(c, a);
        n.crossVectors(u, v);
        w.subVectors(point, a);

        u.cross(w);
        var r = u.dot(n) / n.dot(n);
        w.cross(v);
        var b = w.dot(n) / n.dot(n);
        var a = 1 - r - b;

        if (a >= 0 && a <= 1 && b >= 0 && b <= 1 && r >= 0 && r <= 1) {

            var normal = THREE.Triangle.normal(a, b, c);
            var nearestPoint = this.nearestPointInPointToPlane(point, a, normal);

            return nearestPoint;
        }
    };

    MeasureCommon.nearestPointsInLineSegmentToTriangle = function(p1, p2, a, b, c) {

        // The closest pair of points between a line segment and a triangle can always be found either
        // (a) between an endpoint of the segment an the triangle interior or
        // (b) between the segment and an edge of the triangle.

        var nearestPoints = [];
        var minDist = Number.MAX_VALUE;

        var p3, p4;

        var pp1 = MeasureCommon.pointProjectsInTriangle(p1, a, b, c);
        if (pp1 && p1.distanceTo(pp1) < minDist) {
            minDist = p1.distanceTo(pp1);
            nearestPoints[0] = p1;
            nearestPoints[1] = pp1;
        }

        var pp2 = MeasureCommon.pointProjectsInTriangle(p2, a, b, c);
        if (pp2 && p2.distanceTo(pp2) < minDist) {
            minDist = p2.distanceTo(pp2);
            nearestPoints[0] = p2;
            nearestPoints[1] = pp2;
        }

        p3 = a;
        p4 = b;
        var p = MeasureCommon.nearestPointsInSegmentToSegment(p1, p2, p3, p4);
        if (p[0].distanceTo(p[1]) < minDist) {
            minDist = p[0].distanceTo(p[1]);
            nearestPoints[0] = p[0].clone();
            nearestPoints[1] = p[1].clone();
        }

        p3 = a;
        p4 = c;
        p = MeasureCommon.nearestPointsInSegmentToSegment(p1, p2, p3, p4);
        if (p[0].distanceTo(p[1]) < minDist) {
            minDist = p[0].distanceTo(p[1]);
            nearestPoints[0] = p[0].clone();
            nearestPoints[1] = p[1].clone();
        }

        p3 = b;
        p4 = c;
        p = MeasureCommon.nearestPointsInSegmentToSegment(p1, p2, p3, p4);
        if (p[0].distanceTo(p[1]) < minDist) {

            nearestPoints[0] = p[0].clone();
            nearestPoints[1] = p[1].clone();
        }

        return nearestPoints;
    };

    // Find the two nearest points between triangle and triangle
    MeasureCommon.nearestPointsInTriangleToTriangle = function (a1, b1, c1, a2, b2, c2) {

        // A pair of closest points between two triangles can be found by computing the closest points between
        // segment and triangle for all six possible combinations of an edge from one triangle tested against
        // the other triangle. But segment-triangle distance tests are fairly expensive,  and thus a better
        // realization is that the closest pair of points between T1 and T2 can be found to occur either on
        // an edge from each triangle or as a vertex of one triangle and a point interior to the other triangle.
        // In all, six vertex-triangle tests and nine edge-edge tests are required.

        var nearestPoints = [];
        var minDist = Number.MAX_VALUE;

        function vertexToTriangleTest(point, a, b, c) {

            var p = MeasureCommon.pointProjectsInTriangle(point, a, b, c);
            if (p && point.distanceTo(p) < minDist) {
                minDist = point.distanceTo(p);
                nearestPoints[0] = point;
                nearestPoints[1] = p;
            }
        }

        function edgeToEdgeTest(p1, p2, p3, p4) {

            var p = MeasureCommon.nearestPointsInSegmentToSegment(p1, p2, p3, p4);
            if (p[0].distanceTo(p[1]) < minDist) {
                minDist = p[0].distanceTo(p[1]);
                nearestPoints = p;
            }
        }

        // a1
        vertexToTriangleTest(a1, a2, b2, c2);

        // b1
        vertexToTriangleTest(b1, a2, b2, c2);

        // c1
        vertexToTriangleTest(c1, a2, b2, c2);

        // a2
        vertexToTriangleTest(a2, a1, b1, c1);

        // b2
        vertexToTriangleTest(b2, a1, b1, c1);

        // c2
        vertexToTriangleTest(c2, a1, b1, c1);

        // edge a1b1 and a2b2
        edgeToEdgeTest(a1, b1, a2, b2);

        // edge a1b1 and a2c2
        edgeToEdgeTest(a1, b1, a2, c2);

        // edge a1b1 and b2c2
        edgeToEdgeTest(a1, b1, b2, c2);

        // edge a1c1 and a2b2
        edgeToEdgeTest(a1, c1, a2, b2);

        // edge a1c1 and a2c2
        edgeToEdgeTest(a1, c1, a2, c2);

        // edge a1c1 and b2c2
        edgeToEdgeTest(a1, c1, b2, c2);

        // edge b1c1 and a2b2
        edgeToEdgeTest(b1, c1, a2, b2);

        // edge b1c1 and a2c2
        edgeToEdgeTest(b1, c1, a2, c2);

        // edge b1c1 and b2c2
        edgeToEdgeTest(b1, c1, b2, c2);

        return nearestPoints;
    };

    // Find the two nearest points between edge and face
    MeasureCommon.nearestPointsInEdgeToFace = function(p1, p2, face) {

        var nearestPoints = [];
        var minDist = Number.MAX_VALUE;

        for (var i = 0; i < face.vertices.length; i += 3) {

            var tempPs = MeasureCommon.nearestPointsInLineSegmentToTriangle(p1, p2, face.vertices[i], face.vertices[i + 1], face.vertices[i + 2]);
            if (tempPs[0].distanceTo(tempPs[1]) < minDist) {
                minDist = tempPs[0].distanceTo(tempPs[1]);
                nearestPoints = tempPs;
            }
        }

        return nearestPoints;
    };

    // Get the angle between line and plane
    MeasureCommon.angleLineToPlane = function(v, n) {

        var angle = MeasureCommon.angleVectorToVector(v, n);

        if (angle > 90) {
            angle -= 90;
        }
        else {
            angle = 90 - angle;
        }

        return angle;
    };

    // Get the intersect point between line and plane
    MeasureCommon.intersectPointLineToPlane = function(p0, n0, p1, p2) {

        var u = new THREE.Vector3();
        var w = new THREE.Vector3();
        u.subVectors(p2, p1);
        w.subVectors(p1, p0);

        var D = n0.dot(u);
        var N = -n0.dot(w);

        if (Math.abs(D) < avem.EPSILON) {  // edge is parallel to plane
            if (N == 0)                      // edge lies in plane
                return null;
            else
                return null;                    // no intersection
        }

        // they are not parallel
        u.multiplyScalar(N / D);             // compute segment intersect point
        u.add(p1);
        return u;
    };

    // Find the vertex need to draw For circular arc's radius
    MeasureCommon.nearestVertexInVertexToEdge = function(vertex, edge) {

        var nearestPoint;
        var minDist = Number.MAX_VALUE;

        for (var i = 0; i < edge.vertices.length; i++) {
            var dist = vertex.distanceTo(edge.vertices[i]);
            if (minDist > dist) {
                nearestPoint = edge.vertices[i];
                minDist = dist;
            }
        }

        return nearestPoint;
    };

    MeasureCommon.distancePointToLine = function (point, lineStart, lineEnd) {

        var X0 = new THREE.Vector3();
        var X1 = new THREE.Vector3();
        var distance;
        var param;

        X0.subVectors(lineStart, point);
        X1.subVectors(lineEnd, lineStart);
        param = X0.dot(X1);
        X0.subVectors(lineEnd, lineStart);
        param = -param / X0.dot(X0);

        if (param < 0) {
            distance = point.distanceTo(lineStart);
        }
        else if (param > 1) {
            distance = point.distanceTo(lineEnd);
        }
        else {
            X0.subVectors(point, lineStart);
            X1.subVectors(point, lineEnd);
            X0.cross(X1);
            X1.subVectors(lineEnd, lineStart);

            distance = Math.sqrt(X0.dot(X0)) / Math.sqrt(X1.dot(X1));
        }

        return distance;
    };

    // Helper to iterator over all properties in an object
    MeasureCommon.forAll = function(obj, func) {
        if (obj) {
            for (var i in obj) {
                if (obj.hasOwnProperty(i)) {
                    var m = obj[i];
                    if (m)
                        func(m);
                }
            }
        }
    };

    MeasureCommon.createCommonOverlay = function(viewer, name) {
        if (!viewer.impl.overlayScenes[name])
            viewer.impl.createOverlayScene(name);
    };

    MeasureCommon.safeToggle = function(element, property, show) {
        // toggle only if it needs to. Necessary for IE11.

        if ((element.classList.contains(property) && !show) || (!element.classList.contains(property) && show)) {
            element.classList.toggle(property, show);    
        }
    };

    // Centroid of a polygon is the average of its points.
    MeasureCommon.getCentroidOfPolygon = function(points) {
        var centroid = new THREE.Vector3();
        var n = points.length;

        for (var i = 0; i < n; i++) {
            centroid.add(points[i]);
        }

        centroid.multiplyScalar(1 / n);
        return centroid;
    };

    // Algorithm taken from here:
    // https://www.codeproject.com/Tips/862988/Find-the-Intersection-Point-of-Two-Line-Segments
    MeasureCommon.isSegmentsIntersect = function(p1, p2, q1, q2) {
        var r = new THREE.Vector3();
        var s = new THREE.Vector3();
        var rxs = new THREE.Vector3();
        var tmp = new THREE.Vector3();

        r.subVectors(p2, p1);
        s.subVectors(q2,q1);
        rxs.crossVectors(r, s);

        // The two lines are parallel and non-intersecting.
        if (Math.abs(rxs.z) < avem.EPSILON)
            return false;

        var t = tmp.subVectors(q1, p1).cross(s).z / rxs.z;
        var u = tmp.subVectors(q1, p1).cross(r).z / rxs.z;

        // the two line segments meet at the point p + t r = q + u s.
        if (!(Math.abs(rxs.z) < avem.EPSILON) && (0 <= t && t <= 1) && (0 <= u && u <= 1))
        {
            // An intersection was found.
            return true;
        }

        // 5. Otherwise, the two line segments are not parallel but do not intersect.
        return false;
    };

    // A polygon is a simple polygon if there are no intersection between its edges.
    MeasureCommon.isPolygonSimple = function(points) {
        var n = points.length;

        if (n === 3) {
            return true;
        }

        for (var i = 0; i < n; i++) {
            
            var p1 = points[i];
            var p2 = points[(i + 1) % n];

            for (var j = i + 2; j < n; j++) {

                var q1 = points[(j) % n];
                var q2 = points[(j + 1) % n];

                if (q2 === p1) {
                    break;
                }

                if (MeasureCommon.isSegmentsIntersect(p1, p2, q1, q2)) {
                    return false;
                }
            }
        }

        return true;
    };

    // Algorithm based on https://github.com/mapbox/polylabel
    MeasureCommon.getPolygonVisualCenter = function(polygon) {
        function compareMax(a, b) {
            return b.max - a.max;
        }

        function Cell(x, y, h, polygon) {
            this.x = x; // cell center x
            this.y = y; // cell center y
            this.h = h; // half the cell size
            this.d = pointToPolygonDist(x, y, polygon); // distance from cell center to polygon
            this.max = this.d + this.h * Math.SQRT2; // max distance to polygon within a cell
        }

        // signed distance from point to polygon outline (negative if point is outside)
        function pointToPolygonDist(x, y, polygon) {
            var inside = false;
            var minDistSq = Infinity;

            for (var k = 0; k < polygon.length; k++) {
                var ring = polygon[k];

                for (var i = 0, len = ring.length, j = len - 1; i < len; j = i++) {
                    var a = ring[i];
                    var b = ring[j];

                    if ((a.y > y !== b.y > y) &&
                        (x < (b.x - a.x) * (y - a.y) / (b.y - a.y) + a.x)) inside = !inside;

                    minDistSq = Math.min(minDistSq, getSegDistSq(x, y, a, b));
                }
            }

            return (inside ? 1 : -1) * Math.sqrt(minDistSq);
        }

        // get polygon centroid
        function getCentroidCell(polygon) {
            var area = 0;
            var x = 0;
            var y = 0;
            var points = polygon[0];

            for (var i = 0, len = points.length, j = len - 1; i < len; j = i++) {
                var a = points[i];
                var b = points[j];
                var f = a.x * b.y - b.x * a.y;
                x += (a.x + b.x) * f;
                y += (a.y + b.y) * f;
                area += f * 3;
            }
            if (area === 0) return new Cell(points[0].x, points[0].y, 0, polygon);
            return new Cell(x / area, y / area, 0, polygon);
        }

        // get squared distance from a point to a segment
        function getSegDistSq(px, py, a, b) {

            var x = a.x;
            var y = a.y;
            var dx = b.x - x;
            var dy = b.y - y;

            if (dx !== 0 || dy !== 0) {

                var t = ((px - x) * dx + (py - y) * dy) / (dx * dx + dy * dy);

                if (t > 1) {
                    x = b.x;
                    y = b.y;

                } else if (t > 0) {
                    x += dx * t;
                    y += dy * t;
                }
            }

            dx = px - x;
            dy = py - y;

            return dx * dx + dy * dy;
        }

        function TinyQueue(data, compare) {

            function defaultCompare(a, b) {
                return a < b ? -1 : a > b ? 1 : 0;
            }

            if (!(this instanceof TinyQueue)) return new TinyQueue(data, compare);

            this.data = data || [];
            this.length = this.data.length;
            this.compare = compare || defaultCompare;

            if (this.length > 0) {
                for (var i = (this.length >> 1); i >= 0; i--) this._down(i);
            }
        }

        TinyQueue.prototype = {

            push: function (item) {
                this.data.push(item);
                this.length++;
                this._up(this.length - 1);
            },

            pop: function () {
                if (this.length === 0) return undefined;

                var top = this.data[0];
                this.length--;

                if (this.length > 0) {
                    this.data[0] = this.data[this.length];
                    this._down(0);
                }
                this.data.pop();

                return top;
            },

            peek: function () {
                return this.data[0];
            },

            _up: function (pos) {
                var data = this.data;
                var compare = this.compare;
                var item = data[pos];

                while (pos > 0) {
                    var parent = (pos - 1) >> 1;
                    var current = data[parent];
                    if (compare(item, current) >= 0) break;
                    data[pos] = current;
                    pos = parent;
                }

                data[pos] = item;
            },

            _down: function (pos) {
                var data = this.data;
                var compare = this.compare;
                var halfLength = this.length >> 1;
                var item = data[pos];

                while (pos < halfLength) {
                    var left = (pos << 1) + 1;
                    var right = left + 1;
                    var best = data[left];

                    if (right < this.length && compare(data[right], best) < 0) {
                        left = right;
                        best = data[right];
                    }
                    if (compare(best, item) >= 0) break;

                    data[pos] = best;
                    pos = left;
                }

                data[pos] = item;
            }
        };
        
        if (polygon.length === 3) {
            return MeasureCommon.getCentroidOfPolygon(polygon);
        }

        var precision = 0.01;
        polygon = [polygon];

        // find the bounding box of the outer ring
        var minX, minY, maxX, maxY;
        for (var i = 0; i < polygon[0].length; i++) {
            var p = polygon[0][i];
            if (!i || p.x < minX) minX = p.x;
            if (!i || p.y < minY) minY = p.y;
            if (!i || p.x > maxX) maxX = p.x;
            if (!i || p.y > maxY) maxY = p.y;
        }

        var width = maxX - minX;
        var height = maxY - minY;
        var cellSize = Math.min(width, height);
        var h = cellSize / 2;

        // a priority queue of cells in order of their "potential" (max distance to polygon)
        var cellQueue = new TinyQueue(null, compareMax);

        if (cellSize === 0) return [minX, minY];

        // cover polygon with initial cells
        for (var x = minX; x < maxX; x += cellSize) {
            for (var y = minY; y < maxY; y += cellSize) {
                cellQueue.push(new Cell(x + h, y + h, h, polygon));
            }
        }

        // take centroid as the first best guess
        var bestCell = getCentroidCell(polygon);

        // special case for rectangular polygons
        var bboxCell = new Cell(minX + width / 2, minY + height / 2, 0, polygon);
        if (bboxCell.d > bestCell.d) bestCell = bboxCell;

        while (cellQueue.length) {
            // pick the most promising cell from the queue
            var cell = cellQueue.pop();

            // update the best cell if we found a better one
            if (cell.d > bestCell.d) {
                bestCell = cell;
            }

            // do not drill down further if there's no chance of a better solution
            if (cell.max - bestCell.d <= precision) continue;

            // split the cell into four cells
            h = cell.h / 2;
            cellQueue.push(new Cell(cell.x - h, cell.y - h, h, polygon));
            cellQueue.push(new Cell(cell.x + h, cell.y - h, h, polygon));
            cellQueue.push(new Cell(cell.x - h, cell.y + h, h, polygon));
            cellQueue.push(new Cell(cell.x + h, cell.y + h, h, polygon));
        }

        return {x:bestCell.x, y:bestCell.y};
    };

})();