(function(){ 'use strict';

var MeasureCommon = Autodesk.Viewing.Extensions.B3LMeasure.Functions;
var avem = Autodesk.Viewing.Extensions.B3LMeasure;

/**
 * 
 * @private
 */
Autodesk.Viewing.Extensions.B3LMeasure.Measurement = function(measurementType, id) {

    this.measurementType = measurementType;
    this.id = id;
    this.picks = [];
    this.closedArea = false;

    this.resetMeasureValues();
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.resetMeasureValues = function() {

    this.angle = 0;
    this.distanceX = 0;
    this.distanceY = 0;
    this.distanceZ = 0;
    this.distanceXYZ = 0;
    this.result = null;
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.setPick = function(index, value) {
    
    var pick = this.picks[index] = value;
    pick.id = parseInt(index);
    return pick;
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.getPick = function(index) {
    
    var pick = this.picks[index];
    
    if (!pick) {
        pick = this.setPick(index, new Autodesk.Viewing.Extensions.B3LMeasure.SnapResult());
    }

    return pick;
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.countPicks = function() {
    
    return Object.keys(this.picks).length;
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.getMaxNumberOfPicks = function() {
    
    switch (this.measurementType) {
        case avem.MEASUREMENT_POINT:
            return 1;

        case avem.MEASUREMENT_DISTANCE:
            return 2;

        case avem.MEASUREMENT_ANGLE:
            return 3;

        case avem.MEASUREMENT_AREA:
            return this.closedArea ? this.countPicks() : Number.MAX_VALUE - 1;
    }
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.hasPick = function(pickNumber) {
    
    return this.picks[pickNumber] && !this.picks[pickNumber].isEmpty();
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.isComplete = function() {
    var complete = this.countPicks() === this.getMaxNumberOfPicks();

    for (var key in this.picks) {
        if (this.picks.hasOwnProperty(key)) {
            complete = complete && this.hasPick(key);

            if (!complete)
                break;
        }
    }

    return complete;
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.isEmpty = function() {
    var empty = true;

    for (var key in this.picks) {
        if (this.picks.hasOwnProperty(key)) {
            empty = empty && !this.hasPick(key);

            if(!empty)
                break;
        }
    }

    return empty;
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.clearPick = function(pickNumber) {

    if (this.picks[pickNumber]) {
        this.picks[pickNumber].clear();   
    }

    this.resetMeasureValues();
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.clearAllPicks = function() {
    
    for (var key in this.picks) {
        if (this.picks.hasOwnProperty(key)) {
            this.clearPick(key);
        }
    }
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.hasEqualPicks = function(firstPick, secondPick) {
    if (!firstPick || !secondPick)
        return false;

    if (firstPick.geomType === secondPick.geomType) {
        var first = MeasureCommon.getSnapResultPosition(firstPick);
        var second = MeasureCommon.getSnapResultPosition(secondPick);
        return MeasureCommon.isEqualVectors(first, second, avem.EPSILON);
    }

    return false;
};

/**
 * Calculates distance/angle based on the values of the picks
 * and stores it in .result
 */
Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.computeResult = function(picks, viewer) {

    this.resetMeasureValues();

    if (!viewer.model) {
        this.result = null;
        return false;
    }
    
    // Compute and check if there's a result
    var result = this.result = MeasureCommon.computeResult(picks, this.measurementType, viewer);
    
    if (result === null) {
        return !this.isComplete();
    }
    
    switch (result.type) {
        case avem.MEASUREMENT_POINT:
            this.coordinateX = result.X;
            this.coordinateY = result.Y;
            this.coordinateZ = result.Z;
            return true;

        case avem.MEASUREMENT_DISTANCE:
            this.distanceXYZ = result.distanceXYZ;
            this.distanceX = result.distanceX;
            this.distanceY = result.distanceY;
            this.distanceZ = result.distanceZ;
            return true;

        case avem.MEASUREMENT_ANGLE:
            this.angle = isNaN(result.angle) ? 0 : result.angle;
            return true;

        case avem.MEASUREMENT_AREA:
            this.area = result.area;
            return true;
        
        default:
            return false;
    }
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.getGeometry = function (pickNumber) {
    return {"type": this.picks[pickNumber].geomType, "geometry": this.picks[pickNumber].getGeometry()};
};

Autodesk.Viewing.Extensions.B3LMeasure.Measurement.prototype.attachIndicator = function (viewer, tool, indicatorClass) {
    this.indicator = new indicatorClass(viewer, this, tool);
    this.indicator.init();
};

})();

(function(){ 'use strict';

    var av = Autodesk.Viewing;
    var avem = Autodesk.Viewing.Extensions.B3LMeasure;

    // /** @constructor */
    function MeasurementsManager(viewer)
    {   
        this.viewer = viewer;
        this.init();
    }

    var proto = MeasurementsManager.prototype;

    proto.getCurrentMeasurement = function() {
        return this.currentMeasurement;
    };

    proto.selectMeasurementById = function(id) {
        var measurement = this.measurementsList[id];
        
        if (measurement) {
            this.changeCurrentMeasurement(measurement);
            return this.currentMeasurement;    
        }

        return false;
    };

    proto.createMeasurement = function(measurementType) {
        var id = this.measurementsCounter;
        var measurement = new avem.Measurement(measurementType, id);
        this.measurementsList[id] = measurement;
        this.measurementsCounter++;
        this.changeCurrentMeasurement(measurement);
        return this.currentMeasurement;
    };

    proto.changeCurrentMeasurement = function(measurement) {
        this.currentMeasurement = measurement;
        this.viewer.dispatchEvent({ type: avem.MEASUREMENT_CHANGED_EVENT, data: measurement.measurementType });
    };

    proto.removeCurrentMeasurement = function() { 
        // Remove current measurement from the list
        if (Object.keys(this.measurementsList).length > 0) {
            delete this.measurementsList[this.currentMeasurement.id];
        }
    };

    proto.init = function() {
        this.reset();
    };

    proto.destroy = function() {
        this.reset();
    };

    proto.reset = function() {
        this.currentMeasurement = null;
        this.measurementsList = {};
        this.measurementsCounter = 0;
    };

    Autodesk.Viewing.Extensions.B3LMeasure.MeasurementsManager = MeasurementsManager;

})();
